#!/usr/bin/env python3

from trojanzoo.utils.memory import output_memory
from trojanzoo.utils.module import Process
from trojanzoo.utils.output import prints

import torch
from abc import ABC, abstractmethod

from typing import TYPE_CHECKING
from collections.abc import Callable, Iterable    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class Optimizer(ABC, Process):
    r"""An abstract input optimizer class that inherits
    :class:`trojanzoo.utils.module.Process`.

    Args:
        iteration (int): Optimization iteration.
            Defaults to ``20``.
        stop_threshold (float | None): Threshold used in early stop check.
            Defaults to ``None`` (no early stop).
        loss_fn (~collections.abc.Callable):
            Loss function (it's usually ``reduction='none'``).
        **kwargs: Keyword Arguments passed to
            :class:`trojanzoo.utils.module.Process`.
    """
    name: str = 'optimizer'

    def __init__(self, iteration: int = 20,
                 stop_threshold: None | float = None,
                 loss_fn: Callable[..., torch.Tensor] = None, **kwargs):
        super().__init__(**kwargs)
        self.param_list['optimize'] = ['iteration', 'stop_threshold']

        self.iteration = iteration
        self.stop_threshold = stop_threshold
        self.loss_fn = loss_fn

    # ----------------------Overload---------------------------- #

    def optimize(self, _input: torch.Tensor, *args,
                 iteration: int = None,
                 loss_fn: Callable[..., torch.Tensor] = None,
                 stop_threshold: float = None,
                 output: int | Iterable[str] = None,
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Main optimize method.

        Args:
            _input (torch.Tensor): The batched input tensor to optimize.
            iteration (int | None): Optimization iteration.
                Defaults to :attr:`self.iteration`.
            loss_fn (~collections.abc.Callable):
                Loss function (it's usually ``reduction='none'``).
                Defaults to :attr:`self.loss_fn`.
            stop_threshold (float | None): Threshold used in early stop check.
                ``None`` means using :attr:`self.stop_threshold`.
                Defaults to :attr:`self.stop_threshold`.
            output (int | ~collections.abc.Iterable[str]):
                Output level integer or output items.
                If :class:`int`, call :meth:`get_output_int()`.
                Defaults to :attr:`self.output`.

        Returns:
            (torch.Tensor, torch.Tensor):
                batched adversarial input tensor and batched optimization iterations
                (``-1`` if not reaching :attr:`self.threshold`).
        """
        # --------------- Parameter Initialization ------------- #
        iteration = iteration if iteration is not None else self.iteration
        loss_fn = loss_fn or self.loss_fn
        stop_threshold = stop_threshold if stop_threshold is not None \
            else self.stop_threshold
        output = self.get_output(output)
        kwargs.update(iteration=iteration,
                      loss_fn=loss_fn,
                      stop_threshold=stop_threshold,
                      output=output)
        # ------------------------------------------------------ #
        org_input = _input.clone().detach()
        adv_input = _input.clone().detach()
        adv_input = self.preprocess_input(
            *args, adv_input=adv_input,
            org_input=org_input, **kwargs)
        iter_list: torch.Tensor = torch.zeros(len(adv_input), dtype=torch.long)
        current_idx = torch.arange(len(iter_list))
        if 'start' in output:
            self.output_info(*args, org_input=org_input,
                             adv_input=adv_input,
                             mode='start', **kwargs)
        # ------------------------------------------------------ #
        for _iter in range(iteration):
            early_stop_result = self.early_stop_check(
                *args, current_idx=current_idx,
                adv_input=adv_input, org_input=org_input, **kwargs)
            not_early_stop_result = ~early_stop_result
            current_idx = current_idx[not_early_stop_result]
            iter_list[current_idx] += 1
            if early_stop_result.all():
                if 'end' in output:
                    self.output_info(*args, org_input=org_input,
                                     adv_input=adv_input,
                                     mode='end', **kwargs)
                return adv_input.detach(), iter_list
            self.update_input(
                *args, current_idx=current_idx,
                adv_input=adv_input, org_input=org_input, **kwargs)
            if 'middle' in output:
                self.output_info(
                    *args, org_input=org_input, adv_input=adv_input,
                    mode='middle', _iter=_iter, **kwargs)
        early_stop_result = self.early_stop_check(
            *args, current_idx=current_idx,
            adv_input=adv_input, org_input=org_input, **kwargs)
        current_idx = current_idx[~early_stop_result]
        iter_list[current_idx] += 1
        if 'end' in output:
            self.output_info(*args, org_input=org_input, adv_input=adv_input,
                             mode='end', **kwargs)
        return (adv_input.detach(),
                torch.where(iter_list <= iteration, iter_list,
                            -torch.ones_like(iter_list)))

    @torch.no_grad()
    def early_stop_check(self, *args, current_idx: torch.Tensor = None,
                         adv_input: torch.Tensor = None,
                         loss_values: torch.Tensor = None,
                         loss_fn: Callable[[torch.Tensor],
                                           torch.Tensor] = None,
                         stop_threshold: float = None,
                         loss_kwargs: dict[str, torch.Tensor] = {},
                         **kwargs) -> torch.Tensor:
        r"""Early stop check using :attr:`stop_threshold`.

        Args:
            current_idx (torch.Tensor):
                The indices of :attr:`adv_input` need to check
                (Other indices have early stopped).
            adv_input (torch.Tensor):
                The entire batched adversairl input tensor
                with shape ``(N, *)``.
            loss_values (torch.Tensor):
                Batched loss tensor with shape ``(N)``.
                If ``None``, use :attr:`loss_fn`
                and :attr:`adv_input` to calculate.
                Defaults to ``None``.
            loss_fn (collections.abc.Callable | None):
                Loss function (it's usually ``reduction='none'``).
                Defaults to :attr:`self.loss_fn`.
            stop_threshold (float | None): Threshold used in early stop check.
                ``None`` means using :attr:`self.stop_threshold`.
                Defaults to :attr:`self.stop_threshold`.
            loss_kwargs (dict[str, torch.Tensor]):
                Keyword arguments passed to :attr:`loss_fn`,
                which will also be selected according to :attr:`current_idx`.
            *args: Any positional argument (unused).
            **kwargs: Any keyword argument (unused).

        Returns:
            torch.Tensor: Batched ``torch.BoolTensor`` with shape ``(N)``.
        """
        stop_threshold = stop_threshold if stop_threshold is not None \
            else self.stop_threshold
        if stop_threshold is None:
            return torch.zeros(len(current_idx), dtype=torch.bool)
        if loss_values is None:
            current_loss_kwargs = {k: v[current_idx]
                                   for k, v in loss_kwargs.items()}
            loss_values = loss_fn(
                adv_input[current_idx], **current_loss_kwargs)
        assert loss_values.dim() == 1
        if adv_input[current_idx] is not None:
            assert len(loss_values) == len(current_idx)
        return loss_values.detach().cpu() < stop_threshold

    def output_info(self, *args, mode: str = 'start',
                    _iter: int = 0, iteration: int = 0,
                    output: Iterable[str] = None, indent: int = None,
                    **kwargs):
        r"""Output information.

        Args:
            mode (str): The output mode
                (e.g., ``'start', 'end', 'middle', 'memory'``).
                Should be legal strings in :meth:`get_output_int()`.
                Defaults to ``'start'``.
            _iter (int): Current iteration. Defaults to ``0``.
            iteration (int): Total iteration. Defaults to ``0``.
            output (~collections.abc.Iterable[str]): Output items.
                Defaults to :attr:`self.output`.
            indent (int): The space indent for the entire string.
                Defaults to :attr:`self.indent`.
            *args: Any positional argument (unused).
            **kwargs: Any keyword argument (unused).
        """
        output = output if output is not None else self.output
        indent = indent if indent is not None else self.indent
        if mode in ['start', 'end']:
            prints(f'{self.name} Optimize {mode}', indent=indent)
        elif mode in ['middle']:
            prints(self.output_iter(name=self.name, _iter=_iter,
                                    iteration=iteration),
                   indent=indent + 4)
        if 'memory' in output:
            output_memory(indent=indent + 4)

    @abstractmethod
    def update_input(self, current_idx: torch.Tensor,
                     adv_input: torch.Tensor,
                     org_input: torch.Tensor,
                     *args, **kwargs):
        r"""Optimize input tensor for 1 iteration.

        Args:
            current_idx (torch.Tensor):
                The indices of :attr:`adv_input` need to optimize
                (Other indices have early stopped).
            adv_input (torch.Tensor):
                The entire batched adversairl input tensor
                with shape ``(N, *)``.
            org_input (torch.Tensor):
                The entire batched original input tensor
                with shape ``(N, *)``.
        """
        ...

    def preprocess_input(self, *args,
                         adv_input: torch.Tensor = None,
                         org_input: torch.Tensor = None,
                         **kwargs) -> torch.Tensor:
        r"""Optimize input tensor for 1 iteration.

        Args:
            adv_input (torch.Tensor):
                The entire batched adversairl input tensor
                with shape ``(N, *)``.
            org_input (torch.Tensor):
                The entire batched original input tensor
                with shape ``(N, *)``.
        """
        return adv_input



#!/usr/bin/env python3

from trojanvision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.nn as nn
import torchvision.models
from torchvision.models.shufflenetv2 import (ShuffleNet_V2_X0_5_Weights, ShuffleNet_V2_X1_0_Weights,
                                             ShuffleNet_V2_X1_5_Weights, ShuffleNet_V2_X2_0_Weights)
from collections import OrderedDict
from collections.abc import Callable


class _ShuffleNetV2(_ImageModel):
    def __init__(self, name: str = 'shufflenetv2_x1_0', **kwargs):
        try:
            assert name in ShuffleNetV2.available_models, f'{name=}'
        except Exception:
            raise AssertionError(f'model name should be in {ShuffleNetV2.available_models}')
        super().__init__(**kwargs)
        ModelClass: Callable[..., torchvision.models.ShuffleNetV2] = getattr(
            torchvision.models, name[:16])
        _model = ModelClass(num_classes=self.num_classes)
        module_list: list[nn.Module] = []
        if '_comp' in name:
            conv1: nn.Conv2d = _model.conv1[0]
            _model.conv1[0] = nn.Conv2d(conv1.in_channels, conv1.out_channels, kernel_size=conv1.kernel_size,
                                        stride=1, padding=conv1.padding, dilation=conv1.dilation,
                                        groups=conv1.groups, bias=False)
            module_list.append(('conv1', _model.conv1))
        else:
            module_list.append(('conv1', _model.conv1))
            module_list.append(('maxpool', _model.maxpool))
        module_list.extend([
            ('stage2', _model.stage2),
            ('stage3', _model.stage3),
            ('stage4', _model.stage4),
            ('conv5', _model.conv5),
        ])
        self.features = nn.Sequential(OrderedDict(module_list))
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.fc)
        ]))


class ShuffleNetV2(ImageModel):
    r"""ShuffleNet v2 proposed by Ningning Ma from Megvii in ECCV 2018.

    :Available model names:

        .. code-block:: python3

            {'shufflenetv2', 'shufflenetv2_comp',
             'shufflenetv2_x0_5', 'shufflenetv2_x1_0',
             'shufflenetv2_x1_5', 'shufflenetv2_x2_0',
             'shufflenetv2_x0_5_comp', 'shufflenetv2_x1_0_comp',
             'shufflenetv2_x1_5_comp', 'shufflenetv2_x2_0_comp'}

    See Also:
        * torchvision: :any:`torchvision.models.shufflenet_v2_x0_5`
        * paper: `ShuffleNet V2\: Practical Guidelines for Efficient CNN Architecture Design`_

    Note:
        ``_comp`` reduces the first convolutional layer
        from ``kernel_size=7, stride=2, padding=3``

        to ``kernel_size=3, stride=1, padding=1``,
        and removes the ``maxpool`` layer before block layers.

    .. _ShuffleNet V2\: Practical Guidelines for Efficient CNN Architecture Design:
        https://arxiv.org/abs/1807.11164
    """
    available_models = {'shufflenetv2', 'shufflenetv2_comp',
                        'shufflenetv2_x0_5', 'shufflenetv2_x1_0',
                        'shufflenetv2_x1_5', 'shufflenetv2_x2_0',
                        'shufflenetv2_x0_5_comp', 'shufflenetv2_x1_0_comp',
                        'shufflenetv2_x1_5_comp', 'shufflenetv2_x2_0_comp'}

    weights = {
        'shufflenetv2_x0_5': ShuffleNet_V2_X0_5_Weights,
        'shufflenetv2_x1_0': ShuffleNet_V2_X1_0_Weights,
        'shufflenetv2_x1_5': ShuffleNet_V2_X1_5_Weights,
        'shufflenetv2_x2_0': ShuffleNet_V2_X2_0_Weights,
    }

    def __init__(self, name: str = 'shufflenetv2', layer: str = '_x0_5',
                 model: type[_ShuffleNetV2] = _ShuffleNetV2, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    @classmethod
    def get_name(cls, name: str, layer: str = '') -> str:
        layer = layer if '_x' not in name else ''
        return super().get_name(name, layer=layer)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        _dict = super().get_official_weights(**kwargs)
        new_dict = OrderedDict()
        for i, (key, value) in enumerate(_dict.items()):
            prefix = 'features.' if i < len(_dict) - 2 else 'classifier.'
            new_dict[prefix + key] = value
        return new_dict



#!/usr/bin/env python3

# https://pytorch.org/functorch/nightly/notebooks/neural_tangent_kernels.html

import torch
import torch.nn as nn
from torch.nn.utils import stateless

import functools

from typing import Iterable


def empirical_ntk(module: nn.Module, input1: torch.Tensor, input2: torch.Tensor,
                  parameters: dict[str, nn.Parameter] | Iterable[nn.Parameter] = None,
                  compute='full') -> list[torch.Tensor]:
    einsum_expr: str = ''
    match compute:
        case 'full':
            einsum_expr = 'Naf,Mbf->NMab'   # (N, M, C, C)
        case 'trace':
            einsum_expr = 'Naf,Maf->NM'     # (N, M)
        case 'diagonal':
            einsum_expr = 'Naf,Maf->NMa'    # (N, M, C)
        case _:
            raise ValueError(compute)

    if not isinstance(parameters, dict):
        id_map: dict[torch.Tensor, str] = {
            param.data: name for name, param in module.named_parameters()}
        parameters = {id_map[param.data if isinstance(
            param, torch.nn.Parameter) else param]: param
            for param in parameters}
    if parameters is None:
        parameters = dict(module.named_parameters())
    names, values = zip(*parameters.items())

    def func(*params: torch.Tensor, _input: torch.Tensor = None):
        _output: torch.Tensor = stateless.functional_call(
            module, {n: p for n, p in zip(names, params)}, _input)
        return _output  # (N, C)

    jac1: tuple[torch.Tensor] = torch.autograd.functional.jacobian(
        functools.partial(func, _input=input1), values)
    jac2: tuple[torch.Tensor] = torch.autograd.functional.jacobian(
        functools.partial(func, _input=input2), values)
    jac1 = (j.flatten(2) for j in jac1)  # (N, C, D)
    jac2 = (j.flatten(2) for j in jac2)  # (M, C, D)
    result = torch.stack([torch.einsum(einsum_expr, j1, j2) for j1, j2 in zip(jac1, jac2)]).sum(0)
    return result



#!/usr/bin/env python3

# -- Path configuration ------------------------------------------------

import sys
from os import path
from trojanzoo_sphinx_theme.linkcode import linkcode_helper

sys.path.insert(0, path.abspath('./'))
sys.path.insert(0, path.abspath('../../'))

import trojanzoo as package  # noqa

pkg_name = package.__name__
pkg_file = package.__file__
pkg_version = str(package.__version__)
pkg_location = path.dirname(path.dirname(pkg_file))

# autoapi_dirs = ['../../trojanzoo']

# -- General configuration ------------------------------------------------

project = 'TrojanZoo'
author = 'ain-soph'
copyright = f'2021, {author}'

github_user = author
github_repo = pkg_name
github_version = 'main'

github_url = f'https://github.com/{github_user}/{github_repo}/'
gh_page_url = f'https://{github_user}.github.io/{github_repo}/'

html_baseurl = gh_page_url
html_context = {
    'display_github': True,
    'github_user': github_user,
    'github_repo': github_repo,
    'github_version': github_version,
    "conf_py_path": "/docs/source/",  # Path in the checkout to the docs root
}
html_theme_options = {
    'github_url': github_url,

    'doc_items': {
        'AlpsPlot': '/alpsplot',
        'TrojanZoo': '/trojanzoo',
        'trojanzoo_sphinx_theme': '/trojanzoo_sphinx_theme',
        'base': 'https://github.com/ain-soph/base',
    },

    'logo': 'images/logo/trojanzoo-logo.svg',
    'logo_dark': 'images/logo/trojanzoo-logo-dark.svg',
    'logo_icon': 'images/logo/trojanzoo-logo-icon.svg',
}

# -- Extension configuration ----------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.linkcode',  # viewcode
    'sphinx.ext.napoleon',
    'sphinxcontrib.katex',
    'sphinxcontrib.jquery',
    'sphinx_copybutton',
    # 'autoapi.extension',
]


def linkcode_resolve(domain, info):
    return linkcode_helper(
        domain, info,
        prefix=pkg_location,
        github_url=github_url,
        github_version=github_version)


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'matplotlib': ('https://matplotlib.org/stable', None),
    'numpy': ('https://numpy.org/doc/stable', None),
    'pillow': ('https://pillow.readthedocs.io/en/stable/', None),
    'python': ('https://docs.python.org/3', None),
    'skimage': ('https://scikit-image.org/docs/dev/', None),
    'sklearn': ('https://scikit-learn.org/stable/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
    'torchvision': ('https://pytorch.org/vision/stable/', None),
}

# -- General default configuration ----------------------------------------

needs_sphinx = '7.0.0'
templates_path = ['_templates']
source_suffix = '.rst'  # ['.rst', '.md']
root_doc = 'index'

release = pkg_version
version = release if release.find('a') == -1 else release[:release.find('a')]

language = 'en'
exclude_patterns = []

# -- General default extension configuration ------------------------------

# autodoc options
autodoc_docstring_signature = True
autodoc_inherit_docstrings = False
autodoc_preserve_defaults = True
autodoc_typehints = 'none'

toc_object_entries = True
toc_object_entries_show_parents = "hide"
# autoapi_type = 'python'
# autoapi_generate_api_docs = False

# autosectionlabel options
# autosectionlabel throws warnings if section names are duplicated.
# The following tells autosectionlabel to not throw a warning for
# duplicated section names that are in different documents.
autosectionlabel_prefix_document = True

# katex options
katex_prerender = True

# napoleon options
napoleon_use_ivar = True
napoleon_use_rtype = False

# todo options
# If true, `todo` and `todoList` produce output, else they produce nothing.
# todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

html_theme = 'trojanzoo_sphinx_theme'
html_favicon = 'images/favicon.ico'
html_title = " ".join((project, version, "documentation"))



#!/usr/bin/env python3

from trojanvision.datasets.imagefolder import ImageFolder
from trojanzoo.utils.module import Module


class VGGface2(ImageFolder):
    r"""VGGface2 dataset introduced by Qiong Cao in 2018.
    It inherits :class:`trojanvision.datasets.ImageFolder`.

    See Also:
        * paper: `VGGFace2\: A dataset for recognising faces across pose and age`_
        * website (expired dead link): https://www.robots.ox.ac.uk/~vgg/data/vgg_face2

    Attributes:
        name (str): ``'vggface2'``
        num_classes (int): ``8631`` (Why the papaer claims to have 500 more?)
        data_shape (list[int]): ``[3, 224, 224]``
        valid_set (bool): ``False``

    .. _VGGFace2\: A dataset for recognising faces across pose and age:
        https://arxiv.org/abs/1710.08092
    """
    name: str = 'vggface2'
    num_classes = 8631
    valid_set: bool = False
    # https://docs.qingcloud.com/product/ai/deeplearning/
    url = {'train': 'https://appcenter-deeplearning.sh1a.qingstor.com/dataset/vggface2/vggface2_train.tar.gz',
           'test': 'https://appcenter-deeplearning.sh1a.qingstor.com/dataset/vggface2/vggface2_test.tar.gz'}
    md5 = {'train': '88813c6b15de58afc8fa75ea83361d7f',
           'test': 'bb7a323824d1004e14e00c23974facd3'}
    org_folder_name = {'train': 'train'}


class Sample_VGGface2(VGGface2):

    name: str = 'sample_vggface2'
    num_classes = 20
    url = {}
    org_folder_name = {}

    def initialize_folder(self):
        _dict = Module(vars(self))
        _dict.__delattr__('folder_path')
        vggface2 = VGGface2(**_dict)
        vggface2.sample(child_name=self.name, sample_num=self.num_classes)



#!/usr/bin/env python3

from trojanvision.datasets.imageset import ImageSet

import torchvision.datasets as datasets


class CelebA(ImageSet):
    r"""CelebA dataset.
    It inherits :class:`trojanvision.datasets.ImageSet`.

    See Also:
        * torchvision: :any:`torchvision.datasets.CelebA`
        * paper: `Deep Learning Face Attributes in the Wild`_
        * website: http://mmlab.ie.cuhk.edu.hk/projects/CelebA.html

    Attributes:
        name (str): ``'celeba'``
        num_classes (int): ``10``
        data_shape (list[int]): ``[1, 28, 28]``
        norm_par (dict[str, list[float]]):
            ``{'mean': [0.1307], 'std': [0.3081]}``

    .. _Deep Learning Face Attributes in the Wild:
        https://arxiv.org/abs/1411.7766
    """

    name: str = 'celeba'
    num_classes: int = 10177
    data_shape = [3, 64, 64]

    def __init__(self, norm_par: dict[str, list[int]] = {'mean': [0.5, 0.5, 0.5], 'std': [0.5, 0.5, 0.5]}, **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.CelebA(root=self.folder_path, split='all', download=True)

    def _get_org_dataset(self, mode, **kwargs):
        return datasets.CelebA(root=self.folder_path, split=mode, target_type='identity', **kwargs)



#!/usr/bin/env python3

from trojanvision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
import torchvision.models
from torchvision.models.alexnet import AlexNet_Weights


class _AlexNet(_ImageModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        _model = torchvision.models.alexnet(num_classes=self.num_classes)
        self.features = _model.features
        self.pool = _model.avgpool   # nn.AdaptiveAvgPool2d((6, 6))
        if len(self.classifier) == 1 and \
                isinstance(self.classifier[0], nn.Identity):
            self.classifier = _model.classifier


class AlexNet(ImageModel):
    r"""AlexNet proposed by Alex Krizhevsky from Google in 2014.

    :Available model names:

        .. code-block:: python3

            {'alexnet'}

    See Also:
        * torchvision: :any:`torchvision.models.alexnet`
        * paper: `One weird trick for parallelizing convolutional neural networks`_

    .. code-block:: python3

        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )

        self.pool = nn.AdaptiveAvgPool2d((6, 6))
        self.flatten = nn.Flatten(1)

        self.classifier = nn.Sequential(
            nn.Dropout(),
            nn.Linear(256 * 6 * 6, 4096),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, num_classes),
        )

    .. _One weird trick for parallelizing convolutional neural networks:
        https://arxiv.org/abs/1404.5997
    """
    available_models = {'alexnet'}
    weights = {'alexnet': AlexNet_Weights}

    def __init__(self, name: str = 'alexnet',
                 model: type[_AlexNet] = _AlexNet, **kwargs):
        super().__init__(name=name, model=model, **kwargs)



#!/usr/bin/env python3

from trojanzoo.datasets import Dataset
from trojanzoo.configs import config
from trojanzoo.utils.module import get_name, ModelProcess
from trojanzoo.utils.output import ansi

import torch
import os

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from trojanzoo.configs import Config
from trojanzoo.models import Model
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class Attack(ABC, ModelProcess):
    r"""
    | An abstract class representing an attack.
    | It inherits :class:`trojanzoo.utils.module.ModelProcess`.

    Note:
        This is the implementation of attack.
        For users, please use :func:`create` instead, which is more user-friendly.
    """
    name = 'attack'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        r"""Add attack arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            The concrete attack class may override this method to add more arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('--attack', dest='attack_name')
        group.add_argument('--output', type=int,
                           help='output level (default: 0)')
        group.add_argument('--attack_dir',
                           help='directory to contain attack results')
        return group

    @abstractmethod
    def attack(self, **kwargs):
        r"""Main attack method (need overriding)."""
        ...

    def generate_target(self, _input: torch.Tensor, idx: int = 1,
                        same: bool = False, **kwargs):
        r"""Generate target labels of a batched input based on
        the classification confidence ranking index.

        Args:
            _input (torch.Tensor): The input tensor.
            idx (int): The classification confidence
                rank of target class.
                Defaults to ``1``.
            same (bool): Generate the same label
                for all samples using mod.
                Defaults to ``False``.

        Returns:
            torch.Tensor:
                The generated target label with shape ``(N)``.

        See Also:
            This method calls
            :meth:`trojanzoo.models.Model.generate_target()`.

            The implementation is in
            :func:`trojanzoo.utils.model.generate_target()`.
        """
        # TODO: issue 1 anyway to avoid coding so many arguments?
        return self.model.generate_target(_input, idx=idx, same=same, **kwargs)


def add_argument(parser: argparse.ArgumentParser, attack_name: str = None,
                 attack: str | Attack = None,
                 class_dict: dict[str, type[Attack]] = {}):
    r"""
    | Add attack arguments to argument parser.
    | For specific arguments implementation, see :meth:`Attack.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        attack_name (str): The attack name.
        attack (str | Attack): The attack instance or attack name
            (as the alias of `attack_name`).
        class_dict (dict[str, type[Attack]]):
            Map from attack name to attack class.
            Defaults to ``{}``.

    Returns:
        argparse._ArgumentGroup: The argument group.
    """
    attack_name = get_name(
        name=attack_name, module=attack, arg_list=['--attack'])
    group = parser.add_argument_group(
        '{yellow}attack{reset}'.format(**ansi), description=attack_name)
    try:
        AttackType = class_dict[attack_name]
    except KeyError:
        if attack_name is None:
            print(f'{ansi["red"]}you need to first claim the attack name '
                  f'using "--attack".{ansi["reset"]}')
        print(f'{attack_name} not in \n{list(class_dict.keys())}')
        raise
    return AttackType.add_argument(group)


def create(attack_name: str = None, attack: str | Attack = None,
           dataset_name: str = None, dataset: str | Dataset = None,
           model_name: str = None, model: str | Model = None,
           config: Config = config, class_dict: dict[str, type[Attack]] = {},
           **kwargs) -> Attack:
    r"""
    | Create an attack instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{attack_dir}/{dataset.data_type}/{dataset.name}/{model.name}/{attack.name}'``.
    | For attack implementation, see :class:`Attack`.

    Args:
        attack_name (str): The attack name.
        attack (str | Attack): The attack instance or attack name
            (as the alias of `attack_name`).
        dataset_name (str): The dataset name.
        dataset (str | Dataset):
            Dataset instance or dataset name
            (as the alias of `dataset_name`).
        model_name (str): The model name.
        model (str | Model): The model instance or model name
            (as the alias of `model_name`).
        config (Config): The default parameter config.
        class_dict (dict[str, type[Attack]]):
            Map from attack name to attack class.
            Defaults to ``{}``.
        **kwargs: The keyword arguments
            passed to attack init method.

    Returns:
        Attack: The attack instance.
    """
    dataset_name = get_name(
        name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model,
                          arg_list=['-m', '--model'])
    attack_name = get_name(
        name=attack_name, module=attack, arg_list=['--attack'])
    if dataset_name is None:
        dataset_name = config.full_config['dataset']['default_dataset']
    general_config = config.get_config(dataset_name=dataset_name)['attack']
    specific_config = config.get_config(dataset_name=dataset_name)[attack_name]
    result = general_config.update(specific_config).update(kwargs)
    try:
        AttackType = class_dict[attack_name]
    except KeyError:
        print(f'{attack_name} not in \n{list(class_dict.keys())}')
        raise
    if 'folder_path' not in result.keys():
        folder_path = result['attack_dir']
        if isinstance(dataset, Dataset):
            folder_path = os.path.join(
                folder_path, dataset.data_type, dataset.name)
        if model_name is not None:
            folder_path = os.path.join(folder_path, model_name)
        folder_path = os.path.join(folder_path, AttackType.name)
        result['folder_path'] = folder_path
    return AttackType(name=attack_name, dataset=dataset, model=model, **result)



#!/usr/bin/env python3

# https://github.com/pytorch/examples/blob/main/mnist/main.py

from trojanvision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
from collections import OrderedDict


class _Net(_ImageModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.features = nn.Sequential(OrderedDict([
            ('conv1', nn.Conv2d(1, 32, 3, 1)),
            ('relu1', nn.ReLU(True)),
            ('conv2', nn.Conv2d(32, 64, 3, 1)),
            ('relu2', nn.ReLU(True)),
        ]))
        self.pool = nn.MaxPool2d(2)
        self.classifier = nn.Sequential(OrderedDict([
            ('dropout1', nn.Dropout(0.25)),
            ('fc1', nn.Linear(9216, 128)),
            ('relu1', nn.ReLU(True)),
            ('dropout2', nn.Dropout(0.5)),
            ('fc2', nn.Linear(128, self.num_classes)),
        ]))


class Net(ImageModel):
    available_models = {'net'}

    def __init__(self, name: str = 'net', model: type[_Net] = _Net, **kwargs):
        super().__init__(name=name, model=model, **kwargs)



#!/usr/bin/env python3

from trojanvision.shortcut.pgd import PGD
from trojanvision.datasets import ImageSet
from trojanvision.utils import apply_cmap
from trojanvision.utils.sgm import register_hook
from trojanzoo.models import _Model, Model
from trojanzoo.utils.fim import KFAC, EKFAC
from trojanzoo.utils.tensor import add_noise

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.hooks import RemovableHandle
from torchvision.transforms import Normalize
import re

from typing import TYPE_CHECKING
from trojanzoo.utils.model import ExponentialMovingAverage
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
from matplotlib.colors import Colormap  # type: ignore  # TODO
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data

from matplotlib.cm import get_cmap  # type: ignore  # TODO
jet = get_cmap('jet')


def replace_bn_to_gn(model: nn.Module) -> None:
    r"""Replace all :any:`torch.nn.BatchNorm2d` to :any:`torch.nn.GroupNorm`."""
    for name, module in model.named_children():
        replace_bn_to_gn(module)
        if isinstance(module, nn.BatchNorm2d):
            device = module.weight.device
            gn = nn.GroupNorm(module.num_features,
                              module.num_features,
                              device=device)
            setattr(model, name, gn)


def set_first_layer_channel(model: nn.Module,
                            channel: int = 3,
                            **kwargs):
    r"""Replace the input channel of the first
    :any:`torch.nn.Conv2d` or :any:`torch.nn.Linear`.
    """
    for name, module in model.named_children():
        if len(list(module.children())):
            set_first_layer_channel(module, channel=channel)
            break
        else:
            match module:
                case nn.Conv2d():
                    if module.in_channels == channel:
                        break
                    keys = ['out_channels', 'kernel_size', 'bias', 'stride', 'padding']
                    args = {key: getattr(module, key) for key in keys}
                    args['device'] = module.weight.device
                    args.update(kwargs)
                    new_conv = nn.Conv2d(in_channels=channel, **args)
                    setattr(model, name, new_conv)
                case nn.Linear():
                    if module.in_features == channel:
                        break
                    keys = ['out_features', 'bias']
                    args = {key: getattr(module, key) for key in keys}
                    args['device'] = module.weight.device
                    args.update(kwargs)
                    new_linear = nn.Linear(in_features=channel, **args)
                    setattr(model, name, new_linear)
        break


class _ImageModel(_Model):
    def __init__(self, norm_par: dict[str, list[float]] = None,
                 num_classes: int = 1000, **kwargs):
        super().__init__(num_classes=num_classes, norm_par=norm_par, **kwargs)

    @classmethod
    def define_preprocess(cls, norm_par: dict[str, list[float]] = None,
                          **kwargs) -> nn.Module:
        r"""Define preprocess before feature extractor.

        Args:
            norm_par (dict[str, list[float]]): Normalize parameters.
                It should contain keys of ``['mean', 'std']``.
                Defaults to ``None`` (:any:`torch.nn.Identity`).
            **kwargs: Any keyword argument (unused).

        Returns:
            torchvision.transforms.Normalize | torch.nn.Identity:

        :Example:
            >>> from trojanvision.models import _ImageModel
            >>>
            >>> _ImageModel.define_preprocess()
            Identity()
            >>> norm_par = {'mean':[0.0, 0.0, 0.0], 'std':[1.0, 1.0, 1.0]}
            >>> _ImageModel.define_preprocess(norm_par=norm_par)
            Normalize(mean=[0.0, 0.0, 0.0], std=[1.0, 1.0, 1.0])

        """
        if norm_par is not None:
            return Normalize(mean=norm_par['mean'],
                             std=norm_par['std'])
        return super().define_preprocess(**kwargs)


class ImageModel(Model):
    r"""
    | A basic image model wrapper class, which should be the most common interface for users.
    | It inherits :class:`trojanzoo.models.Model` and further extend
      adversarial training and Skip Gradient Method (SGM).

    See Also:
        * Adversarial Training:

          - Free: https://github.com/mahyarnajibi/FreeAdversarialTraining
          - Fast: https://github.com/locuslab/fast_adversarial
        * Skip Gradient Method: https://github.com/csdongxian/skip-connections-matter

    Args:
        layer (int | str): Default layer when it's not provided in :attr:`name`.
            Defaults to ``''``.

    Attributes:
        pgd (trojanvision.attacks.PGD):
            PGD attacker using eval settings without early stop.
            It's only constructed when :attr:`adv_train` is not ``None``.
        adv_train (str | None): Adversarial training strategy.
            Choose from ``[None, 'pgd', 'free', 'trades']``.
            Defaults to ``None``.

            Note:
                If ``adv_train is not None and suffix is None``,
                set ``suffix = f'_at-{adv_train}'``.
        adv_train_random_init (bool): Whether to random initialize adversarial noise
            using normal distribution with :attr:`adv_train_eps`.
            Otherwise, attack starts from the benign inputs.
            Defaults to ``False``.
        adv_train_iter (int): Adversarial training PGD iteration.
            Defaults to ``7``.
        adv_train_alpha (float): Adversarial training PGD alpha.
            Defaults to :math:`\frac{2}{255}`.
        adv_train_eps (float): Adversarial training PGD eps.
            Defaults to :math:`\frac{8}{255}`.
        adv_train_eval_iter (int): Adversarial training PGD iteration at evaluation.
            Defaults to :attr:`adv_train_iter`.
        adv_train_eval_alpha (float): Adversarial training PGD alpha at evaluation.
            Defaults to :attr:`adv_train_alpha`.
        adv_train_eval_eps (float): Adversarial training PGD eps at evaluation.
            Defaults to :attr:`adv_train_eps`.
        adv_train_trades_beta (float): regularization factor
            (:math:`\frac{1}{\lambda}` in TRADES)
            Defaults to ``6.0``.
        norm_layer (str): The normalization layer type.
            Choose from ``['bn', 'gn']``.
            Defaults to ``['bn']``.
        sgm (bool): Whether to use Skip Gradient Method. Defaults to ``False``.
        sgm_gamma (float): The gradient factor :math:`\gamma` used in SGM.
            Defaults to ``1.0``.
    """

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        r"""Add image model arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            The concrete model class may override this method to add more arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.

        See Also:
            :meth:`trojanzoo.models.Model.add_argument()`
        """
        super().add_argument(group)
        group.add_argument('--adv_train', choices=[None, 'pgd', 'free', 'trades'],
                           help='adversarial training (default: None)')
        group.add_argument('--adv_train_random_init', action='store_true')
        group.add_argument('--adv_train_iter', type=int,
                           help='adversarial training PGD iteration (default: 7).')
        group.add_argument('--adv_train_alpha', type=float,
                           help='adversarial training PGD alpha (default: 2/255).')
        group.add_argument('--adv_train_eps', type=float,
                           help='adversarial training PGD eps (default: 8/255).')
        group.add_argument('--adv_train_eval_iter', type=int)
        group.add_argument('--adv_train_eval_alpha', type=float)
        group.add_argument('--adv_train_eval_eps', type=float)
        group.add_argument('--adv_train_trades_beta', type=float,
                           help='regularization, i.e., 1/lambda in TRADES '
                           '(default: 6.0)')

        group.add_argument('--norm_layer', choices=['bn', 'gn'], default='bn')
        group.add_argument('--sgm', action='store_true',
                           help='whether to use sgm gradient (default: False)')
        group.add_argument('--sgm_gamma', type=float,
                           help='sgm gamma (default: 1.0)')
        return group

    def __init__(self, name: str = 'imagemodel', layer: int | str = '',
                 model: type[_ImageModel] | _ImageModel = _ImageModel,
                 dataset: ImageSet = None, data_shape: list[int] = None,
                 adv_train: str = None, adv_train_random_init: bool = False, adv_train_eval_random_init: bool = None,
                 adv_train_iter: int = 7, adv_train_alpha: float = 2 / 255, adv_train_eps: float = 8 / 255,
                 adv_train_eval_iter: int = None, adv_train_eval_alpha: float = None, adv_train_eval_eps: float = None,
                 adv_train_trades_beta: float = 6.0,
                 norm_layer: str = 'bn', sgm: bool = False, sgm_gamma: float = 1.0,
                 norm_par: dict[str, list[float]] = None, suffix: str = None,
                 modify_first_layer_channel: bool = True, **kwargs):
        self.norm_par = norm_par
        name = self.get_name(name, layer=layer)
        if norm_par is None and isinstance(dataset, ImageSet):
            norm_par = None if dataset.normalize else dataset.norm_par
        if 'num_classes' not in kwargs.keys() and dataset is None:
            kwargs['num_classes'] = 1000
        if adv_train is not None and suffix is None:
            suffix = f'_at-{adv_train}'
        super().__init__(name=name, model=model, dataset=dataset, data_shape=data_shape,
                         norm_par=norm_par, suffix=suffix, **kwargs)
        assert norm_layer in ['bn', 'gn']
        if norm_layer == 'gn':
            replace_bn_to_gn(self._model)

        if data_shape is None:
            assert isinstance(dataset, ImageSet), 'Please specify data_shape or dataset'
            data_shape = dataset.data_shape
        args = {'padding': 3} if 'vgg' in name else {}  # TODO: so ugly
        if modify_first_layer_channel:
            set_first_layer_channel(self._model.features,
                                    channel=data_shape[0], **args)

        self.sgm: bool = sgm
        self.sgm_gamma: float = sgm_gamma
        self.adv_train = adv_train
        self.adv_train_random_init = adv_train_random_init
        self.adv_train_eval_random_init = adv_train_eval_random_init \
            if adv_train_eval_random_init is not None else adv_train_random_init
        self.adv_train_iter = adv_train_iter
        self.adv_train_alpha = adv_train_alpha
        self.adv_train_eps = adv_train_eps
        self.adv_train_eval_iter = adv_train_eval_iter if adv_train_eval_iter is not None else adv_train_iter
        self.adv_train_eval_alpha = adv_train_eval_alpha if adv_train_eval_alpha is not None else adv_train_alpha
        self.adv_train_eval_eps = adv_train_eval_eps if adv_train_eval_eps is not None else adv_train_eps
        self.adv_train_trades_beta = adv_train_trades_beta

        self.param_list['imagemodel'] = []
        if sgm:
            self.param_list['imagemodel'].append('sgm_gamma')
            register_hook(self, sgm_gamma)
        if adv_train is not None:
            if 'suffix' not in self.param_list['model']:
                self.param_list['model'].append('suffix')
            self.param_list['adv_train'] = ['adv_train', 'adv_train_random_init', 'adv_train_eval_random_init',
                                            'adv_train_iter', 'adv_train_alpha', 'adv_train_eps',
                                            'adv_train_eval_iter', 'adv_train_eval_alpha', 'adv_train_eval_eps']
            if adv_train == 'trades':
                self.param_list['adv_train'].append('adv_train_trades_beta')
            self.pgd = PGD(pgd_alpha=self.adv_train_eval_alpha, pgd_eps=self.adv_train_eval_eps,
                           iteration=self.adv_train_eval_iter, stop_threshold=None,
                           target_idx=0,
                           random_init=self.adv_train_eval_random_init,
                           model=self, dataset=self.dataset)
        self._model: _ImageModel
        self.dataset: ImageSet
        self.sgm_remove: list[RemovableHandle]

    def trades_loss_fn(self, _input: torch.Tensor, org_prob: torch.Tensor, **kwargs) -> torch.Tensor:
        return -F.kl_div(self(_input, **kwargs).log_softmax(1), org_prob,
                         reduction='batchmean')

    @classmethod
    def get_name(cls, name: str, layer: int | str = '') -> str:
        r"""A useful function to combine :attr:`name` and :attr:`layer`.

        Note:
            * If there is already layer claimed in :attr:`name`,
              :attr:`layer` will be ignored.
            * You may override this method for concrete model class
              on demand.

        Args:
            name (str): Model name string.
            layer (int | str): Model layer. Defaults to ``''``.

        :Example:
            >>> from trojanvision.models import ImageModel
            >>>
            >>> ImageModel.get_name('vgg_comp', layer=13)
            'vgg13_comp'
            >>> ImageModel.get_name('vgg16_comp', layer=13)
            'vgg16_comp'
        """
        full_list = name.split('_')
        partial_name = full_list[0]
        re_list = re.findall(r'\d+|\D+', partial_name)
        if len(re_list) > 1:
            layer = re_list[1]
        else:
            partial_name += str(layer)
        full_list[0] = partial_name
        return '_'.join(full_list)

    def get_heatmap(self, _input: torch.Tensor, _label: torch.Tensor,
                    method: str = 'grad_cam', cmap: Colormap = jet,
                    mode: str = 'bicubic') -> torch.Tensor:
        r"""Use colormap :attr:`cmap` to get heatmap tensor of :attr:`_input`
        w.r.t. :attr:`_label` with :attr:`method`.

        Args:
            _input (torch.Tensor): The (batched) input tensor
                with shape ``([N], C, H, W)``.
            _label (torch.Tensor): The (batched) label tensor
                with shape ``([N])``
            method (str): The method to calculate heatmap.
                Choose from ``['grad_cam', 'saliency_map']``.
                Defaults to ``'grad_cam'``.
            cmap (matplotlib.colors.Colormap): The colormap to use.
            mode (str): Passed to :any:`torch.nn.functional.interpolate`.
                Defaults to ``'bicubic'``.

        Returns:
            torch.Tensor: The heatmap tensor with shape ([N], C, H, W).

        Note:
            Most :any:`matplotlib.colors.Colormap` will return
            a 4-channel heatmap with alpha channel.

        See Also:
            https://keras.io/examples/vision/grad_cam/

        :Example:
            .. code-block:: python
                :emphasize-lines: 30-32

                import trojanvision
                from trojanvision.utils import superimpose
                import torchvision
                import torchvision.transforms as transforms
                import PIL.Image as Image
                import os
                import wget

                env = trojanvision.environ.create(device='cpu')
                model = trojanvision.models.create(
                    'resnet152', data_shape=[3, 224, 224], official=True,
                    norm_par={'mean': [0.485, 0.456, 0.406],
                              'std': [0.229, 0.224, 0.225]})
                transform = transforms.Compose([
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                    transforms.PILToTensor(),
                    transforms.ConvertImageDtype(torch.float)])
                url = 'https://i.imgur.com/Bvro0YD.png'
                if not os.path.isfile('african_elephant.png'):
                    wget.download(url, 'african_elephant.png')

                img = Image.open('african_elephant.png').convert(mode='RGB')

                _input = transform(img).unsqueeze(0).to(env['device'])
                _prob = model.get_prob(_input).squeeze()
                label = _prob.argmax().item()
                conf = _prob[label].item()
                print(f'{label=:}  {conf=:.2%}')

                grad_cam = model.get_heatmap(_input, label)[:, :3]
                saliency_map = model.get_heatmap(_input, label,
                                                 method='saliency_map')[:, :3]
                grad_cam_impose = (grad_cam * 0.4 + _input)
                saliency_map_impose = (saliency_map * 0.4 + _input)
                grad_cam_impose = grad_cam_impose.div(grad_cam_impose.max())
                saliency_map_impose = saliency_map_impose.div(saliency_map_impose.max())

                torchvision.utils.save_image(_input, './center_cropped.png')
                torchvision.utils.save_image(grad_cam, './grad_cam.png')
                torchvision.utils.save_image(saliency_map, './saliency_map.png')
                torchvision.utils.save_image(grad_cam_impose, './grad_cam_impose.png')
                torchvision.utils.save_image(saliency_map_impose, './saliency_map_impose.png')

            ``label=386  conf=77.74%``

            .. table::
                :align: left
                :widths: 80, 160, 160

                +-----------------------+------------------+------------------------+
                |  original             |  |original|      | |center_cropped|       |
                +-----------------------+------------------+------------------------+
                |  grad_cam             |  |grad_cam|      | |grad_cam_impose|      |
                +-----------------------+------------------+------------------------+
                |  saliency_map         |  |saliency_map|  | |saliency_map_impose|  |
                +-----------------------+------------------+------------------------+

            .. |original| image:: https://i.imgur.com/Bvro0YD.png
                :width: 224px
            .. |center_cropped| image:: /images/trojanvision/center_cropped.png
            .. |grad_cam| image:: /images/trojanvision/grad_cam.png
            .. |grad_cam_impose| image:: /images/trojanvision/grad_cam_impose.png
            .. |saliency_map| image:: /images/trojanvision/saliency_map.png
            .. |saliency_map_impose| image:: /images/trojanvision/saliency_map_impose.png
        """
        squeeze_flag = False
        if _input.dim() == 3:
            _input = _input.unsqueeze(0)    # (N, C, H, W)
            squeeze_flag = True
        if isinstance(_label, int):
            _label = [_label] * len(_input)
        _label = torch.as_tensor(_label, device=_input.device)
        heatmap = _input    # linting purpose
        match method:
            case 'grad_cam':
                feats = self._model.get_fm(_input).detach()   # (N, C', H', W')
                feats.requires_grad_()
                _output: torch.Tensor = self._model.pool(feats)   # (N, C', 1, 1)
                _output = self._model.flatten(_output)   # (N, C')
                _output = self._model.classifier(_output)   # (N, num_classes)
                _output = _output.gather(dim=1, index=_label.unsqueeze(1)).sum()
                grad = torch.autograd.grad(_output, feats)[0]   # (N, C', H', W')
                feats.requires_grad_(False)
                weights = grad.mean(dim=-2, keepdim=True).mean(dim=-1, keepdim=True)    # (N, C', 1, 1)
                heatmap = (feats * weights).sum(dim=1, keepdim=True).clamp(0)  # (N, 1, H', W')
                # heatmap.sub_(heatmap.amin(dim=-2, keepdim=True).amin(dim=-1, keepdim=True))
                heatmap.div_(heatmap.amax(dim=-2, keepdim=True).amax(dim=-1, keepdim=True))
                heatmap: torch.Tensor = F.interpolate(heatmap, _input.shape[-2:], mode=mode)[:, 0]   # (N, H, W)
                # Note that we violate the image order convension (W, H, C)
            case 'saliency_map':
                _input.requires_grad_()
                _output = self(_input).gather(dim=1, index=_label.unsqueeze(1)).sum()
                grad = torch.autograd.grad(_output, _input)[0]   # (N, C, H, W)
                _input.requires_grad_(False)

                heatmap = grad.abs().amax(dim=1)   # (N, H, W)
                heatmap.sub_(heatmap.amin(dim=-2, keepdim=True).amin(dim=-1, keepdim=True))
                heatmap.div_(heatmap.amax(dim=-2, keepdim=True).amax(dim=-1, keepdim=True))
            case _:
                raise NotImplementedError(f'{method=} is not supported yet.')
        heatmap = apply_cmap(heatmap.detach().cpu(), cmap)
        return heatmap[0] if squeeze_flag else heatmap

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor],
                 adv_train: bool = False,
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        r"""
        Args:
            data (tuple[torch.Tensor, torch.Tensor]):
                Tuple of ``(input, label)``.
            adv_train (bool): Whether to Defaults to ``False``.
            **kwargs: Keyword arguments passed to
                :meth:`trojanzoo.models.Model.get_data()`.
        """
        # In training process, `adv_train` args will not be passed to `get_data`.
        # It's passed to `_validate`.
        # So it's always `False`.
        _input, _label = super().get_data(data, **kwargs)
        if adv_train:
            assert self.pgd is not None
            adv_x, _ = self.pgd.optimize(_input=_input, target=_label)
            return adv_x, _label
        return _input, _label

    def _validate(self, adv_train: None | bool | str = None, **kwargs) -> tuple[float, float]:
        r""""""
        adv_train = bool(adv_train) if adv_train is not None else bool(self.adv_train)
        if not adv_train:
            return super()._validate(**kwargs)
        clean_acc, _ = super()._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                         adv_train=False, **kwargs)
        adv_acc, _ = super()._validate(print_prefix='Validate Adv', main_tag='valid adv',
                                       adv_train=True, **kwargs)
        return clean_acc + adv_acc, adv_acc

    def _train(self, epochs: int, optimizer: Optimizer, lr_scheduler: _LRScheduler = None,
               adv_train: bool = None,
               lr_warmup_epochs: int = 0,
               model_ema: ExponentialMovingAverage = None,
               model_ema_steps: int = 32,
               grad_clip: float = None, pre_conditioner: None | KFAC | EKFAC = None,
               print_prefix: str = 'Train', start_epoch: int = 0, resume: int = 0,
               validate_interval: int = 10, save: bool = False, amp: bool = False,
               loader_train: torch.utils.data.DataLoader = None,
               loader_valid: torch.utils.data.DataLoader = None,
               epoch_fn: Callable[..., None] = None,
               get_data_fn: Callable[...,
                                     tuple[torch.Tensor, torch.Tensor]] = None,
               loss_fn: Callable[..., torch.Tensor] = None,
               after_loss_fn: Callable[..., None] = None,
               validate_fn: Callable[..., tuple[float, float]] = None,
               save_fn: Callable[..., None] = None, file_path: str = None,
               folder_path: str = None, suffix: str = None,
               writer=None, main_tag: str = 'train', tag: str = '',
               accuracy_fn: Callable[..., list[float]] = None,
               verbose: bool = True, indent: int = 0, **kwargs):
        adv_train = adv_train if adv_train is not None else bool(self.adv_train)
        if adv_train:
            after_loss_fn_old = after_loss_fn
            if not callable(after_loss_fn) and hasattr(self, 'after_loss_fn'):
                after_loss_fn_old = getattr(self, 'after_loss_fn')
            loss_fn = loss_fn if callable(loss_fn) else self.loss

            def after_loss_fn_new(_input: torch.Tensor, _label: torch.Tensor, _output: torch.Tensor,
                                  loss: torch.Tensor, optimizer: Optimizer, loss_fn: Callable[..., torch.Tensor] = None,
                                  amp: bool = False, scaler: torch.cuda.amp.GradScaler = None, **kwargs):
                optimizer.zero_grad()
                self.zero_grad()
                if pre_conditioner is not None:
                    pre_conditioner.reset()

                if self.adv_train == 'free':
                    noise = self.pgd.init_noise(_input.shape, pgd_eps=self.adv_train_eps,
                                                random_init=self.adv_train_random_init,
                                                device=_input.device)
                    adv_x = add_noise(x=_input, noise=noise, universal=self.pgd.universal,
                                      clip_min=self.pgd.clip_min, clip_max=self.pgd.clip_max)
                    noise.data = self.pgd.valid_noise(adv_x, _input)
                    for m in range(self.adv_train_iter):
                        loss = loss_fn(adv_x, _label)
                        if amp:
                            scaler.scale(loss).backward()
                            scaler.step(optimizer)
                            scaler.update()
                        else:
                            loss.backward()
                            optimizer.step()
                        optimizer.zero_grad()
                        self.zero_grad()
                        # self.eval()
                        adv_x, _ = self.pgd.optimize(_input=_input, noise=noise, target=_label, iteration=1,
                                                     pgd_alpha=self.adv_train_alpha, pgd_eps=self.adv_train_eps)
                        # self.train()
                        loss = loss_fn(adv_x, _label)
                else:
                    loss = self.adv_loss(_input=_input, _label=_label, loss_fn=loss_fn)

                if amp:
                    scaler.scale(loss).backward()
                else:
                    loss.backward()
                if callable(after_loss_fn_old):
                    after_loss_fn_old(_input=_input, _label=_label, _output=_output,
                                      loss=loss, optimizer=optimizer, loss_fn=loss_fn,
                                      amp=amp, scaler=scaler, **kwargs)
            after_loss_fn = after_loss_fn_new

        return super()._train(epochs=epochs, optimizer=optimizer, lr_scheduler=lr_scheduler,
                              adv_train=adv_train,
                              lr_warmup_epochs=lr_warmup_epochs,
                              model_ema=model_ema, model_ema_steps=model_ema_steps,
                              grad_clip=grad_clip, pre_conditioner=pre_conditioner,
                              print_prefix=print_prefix, start_epoch=start_epoch,
                              resume=resume, validate_interval=validate_interval,
                              save=save, amp=amp,
                              loader_train=loader_train, loader_valid=loader_valid,
                              epoch_fn=epoch_fn, get_data_fn=get_data_fn,
                              loss_fn=loss_fn, after_loss_fn=after_loss_fn,
                              validate_fn=validate_fn,
                              save_fn=save_fn, file_path=file_path,
                              folder_path=folder_path, suffix=suffix,
                              writer=writer, main_tag=main_tag, tag=tag,
                              accuracy_fn=accuracy_fn,
                              verbose=verbose, indent=indent, **kwargs)

    def adv_loss(self, _input: torch.Tensor, _label: torch.Tensor,
                 loss_fn: Callable[..., torch.Tensor] = None,
                 adv_train: str = None) -> torch.Tensor:
        adv_train = adv_train if adv_train is not None else self.adv_train
        loss_fn = loss_fn if callable(loss_fn) else self.loss
        match adv_train:
            case 'trades':
                noise = 1e-3 * torch.randn_like(_input)
                org_prob = self.get_prob(_input)
                adv_x, _ = self.pgd.optimize(_input=_input, noise=noise, target=_label,
                                             iteration=self.adv_train_iter,
                                             pgd_alpha=self.adv_train_alpha,
                                             pgd_eps=self.adv_train_eps,
                                             loss_fn=self.trades_loss_fn,
                                             loss_kwargs={'org_prob': org_prob.detach()})
                adv_x = _input + (adv_x - _input).detach()
                return loss_fn(_input, _label) - self.adv_train_trades_beta * \
                    self.trades_loss_fn(_input=adv_x, org_prob=org_prob)
            case 'pgd':
                adv_x, _ = self.pgd.optimize(_input=_input, target=_label,
                                             iteration=self.adv_train_iter,
                                             pgd_alpha=self.adv_train_alpha,
                                             pgd_eps=self.adv_train_eps,
                                             random_init=self.adv_train_random_init)
                adv_x = _input + (adv_x - _input).detach()
                return loss_fn(adv_x, _label)
            case _:
                raise NotImplementedError(f'{adv_train=} is not supported yet.')



#!/usr/bin/env python3

from trojanvision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
import torchvision.models
from torchvision.models.vgg import (VGG11_Weights, VGG11_BN_Weights, VGG13_Weights, VGG13_BN_Weights,
                                    VGG16_Weights, VGG16_BN_Weights, VGG19_Weights, VGG19_BN_Weights)
from collections.abc import Callable


class _VGG(_ImageModel):

    def __init__(self, name: str = 'vgg', dropout: float = 0.5, **kwargs):
        if 'num_features' not in kwargs.keys() and ('_comp' or '_s' in name):
            kwargs['num_features'] = [512] if '_s' in name else [512] * 3
        super().__init__(dropout=dropout, **kwargs)
        class_name = name.replace('_comp', '').replace('_s', '')
        ModelClass: Callable[..., torchvision.models.VGG] = getattr(torchvision.models, class_name)
        _model = ModelClass(num_classes=self.num_classes)
        self.features: nn.Sequential = _model.features
        if '_comp' in name:
            self.pool = nn.AdaptiveAvgPool2d((1, 1))
        else:
            self.pool = _model.avgpool   # nn.AdaptiveAvgPool2d((7, 7))
            self.classifier = _model.classifier
        # nn.Sequential(
        #     nn.Linear(512 * 7 * 7, 4096),
        #     nn.ReLU(True),
        #     nn.Dropout(),
        #     nn.Linear(4096, 4096),
        #     nn.ReLU(True),
        #     nn.Dropout(),
        #     nn.Linear(4096, num_classes),
        # ))


class VGG(ImageModel):
    r"""VGG model proposed by Karen Simonyan from University of Oxford in ICLR 2015.

    :Available model names:

        .. code-block:: python3

            {'vgg', 'vgg_bn', 'vgg_comp', 'vgg_bn_comp', 'vgg_s', 'vgg_bn_s',
             'vgg11', 'vgg13', 'vgg16', 'vgg19',
             'vgg11_bn', 'vgg13_bn', 'vgg16_bn', 'vgg19_bn',
             'vgg11_comp', 'vgg13_comp', 'vgg16_comp', 'vgg19_comp',
             'vgg11_bn_comp', 'vgg13_bn_comp', 'vgg16_bn_comp', 'vgg19_bn_comp'
             'vgg11_s', 'vgg13_s', 'vgg16_s', 'vgg19_s',
             'vgg11_bn_s', 'vgg13_bn_s', 'vgg16_bn_s', 'vgg19_bn_s'}

    See Also:
        * torchvision: :any:`torchvision.models.vgg11`
        * paper: `Very Deep Convolutional Networks for Large-Scale Image Recognition`_

    Note:
        * ``_comp`` sets :any:`torch.nn.AdaptiveAvgPool2d` from ``(7, 7)`` to ``(1, 1)``,
          update the intermediate feature dimension from 4096 to 512 in ``self.classifier``.
        * ``_s`` further makes ``self.classifier`` only one single linear layer based on ``_comp``.

    .. _Very Deep Convolutional Networks for Large-Scale Image Recognition:
        https://arxiv.org/abs/1409.1556
    """
    available_models = {'vgg', 'vgg_bn', 'vgg_comp', 'vgg_bn_comp', 'vgg_s', 'vgg_bn_s',
                        'vgg11', 'vgg13', 'vgg16', 'vgg19',
                        'vgg11_bn', 'vgg13_bn', 'vgg16_bn', 'vgg19_bn',
                        'vgg11_comp', 'vgg13_comp', 'vgg16_comp', 'vgg19_comp',
                        'vgg11_bn_comp', 'vgg13_bn_comp', 'vgg16_bn_comp', 'vgg19_bn_comp'
                        'vgg11_s', 'vgg13_s', 'vgg16_s', 'vgg19_s',
                        'vgg11_bn_s', 'vgg13_bn_s', 'vgg16_bn_s', 'vgg19_bn_s'}

    weights = {
        'vgg11': VGG11_Weights,
        'vgg11_bn': VGG11_BN_Weights,
        'vgg13': VGG13_Weights,
        'vgg13_bn': VGG13_BN_Weights,
        'vgg16': VGG16_Weights,
        'vgg16_bn': VGG16_BN_Weights,
        'vgg19': VGG19_Weights,
        'vgg19_bn': VGG19_BN_Weights,
    }

    def __init__(self, name: str = 'vgg', layer: int = 13,
                 model: type[_VGG] = _VGG, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)



#!/usr/bin/env python3

from trojanvision.datasets.imageset import ImageSet

import torchvision.datasets as datasets


class MNIST(ImageSet):
    r"""MNIST dataset.
    It inherits :class:`trojanvision.datasets.ImageSet`.

    See Also:
        * torchvision: :any:`torchvision.datasets.MNIST`
        * paper: `Gradient-based learning applied to document recognition`_
        * website: http://yann.lecun.com/exdb/mnist/

    Attributes:
        name (str): ``'mnist'``
        num_classes (int): ``10``
        data_shape (list[int]): ``[1, 28, 28]``
        norm_par (dict[str, list[float]]):
            ``{'mean': [0.1307], 'std': [0.3081]}``

    .. _Gradient-based learning applied to document recognition:
        https://ieeexplore.ieee.org/document/726791
    """

    name: str = 'mnist'
    num_classes: int = 10
    data_shape = [1, 28, 28]

    def __init__(self, norm_par: dict[str, list[int]] = {'mean': [0.1307], 'std': [0.3081]}, **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.MNIST(root=self.folder_path, train=True, download=True)
        datasets.MNIST(root=self.folder_path, train=False, download=True)

    def _get_org_dataset(self, mode, **kwargs):
        assert mode in ['train', 'valid']
        return datasets.MNIST(root=self.folder_path, train=(mode == 'train'), **kwargs)



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --epochs 200 --batch_size 96 --cutout --lr 0.025 --lr_scheduler

# adv train pgd
CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --adv_train --adv_train_random_init --validate_interval 1 --epochs 15 --lr 0.1 --lr_scheduler

# adv train fgsm
CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --adv_train --adv_train_random_init --adv_train_iter 1 --adv_train_alpha 0.0392156862745 --adv_train_eval_iter 7 --adv_train_eval_alpha 0.0078431372549 --validate_interval 1 --epochs 15 --lr 0.1 --lr_scheduler

# adv train fgsm mnist
CUDA_VISIBLE_DEVICES=0 python examples/train.py --verbose 1 --color --dataset mnist --adv_train --adv_train_random_init --adv_train_iter 1 --adv_train_alpha 0.375 --adv_train_eps 0.3 --adv_train_eval_iter 7 --adv_train_eval_alpha 0.1 --adv_train_eval_eps 0.3 --validate_interval 1 --epochs 15 --lr 0.1 --lr_scheduler
"""  # noqa: E501

import trojanvision
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    trojanvision.environ.add_argument(parser)
    trojanvision.datasets.add_argument(parser)
    trojanvision.models.add_argument(parser)
    trojanvision.trainer.add_argument(parser)
    kwargs = vars(parser.parse_args())

    env = trojanvision.environ.create(**kwargs)
    dataset = trojanvision.datasets.create(**kwargs)
    model = trojanvision.models.create(dataset=dataset, **kwargs)
    trainer = trojanvision.trainer.create(dataset=dataset, model=model, **kwargs)

    if env['verbose']:
        trojanvision.summary(env=env, dataset=dataset, model=model, trainer=trainer)
    model._train(**trainer)



#!/usr/bin/env python3

# https://github.com/VICO-UoE/DatasetCondensation/blob/master/networks.py

import torch
import torch.nn as nn
from torchvision.models.resnet import conv1x1, conv3x3, ResNet

from typing import Any, Callable


class BasicBlockAP(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: None | nn.Module = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: None | Callable[..., nn.Module] = None
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride=1)    # modification
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)

        self.pool = nn.Identity() if stride == 1 else nn.AvgPool2d(kernel_size=stride, stride=stride)   # modification

        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.pool(out)    # modification

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class BottleneckAP(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: None | nn.Module = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: None | Callable[..., nn.Module] = None
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride=1, groups=groups, dilation=dilation)  # modification
        self.bn2 = norm_layer(width)
        self.pool = nn.Identity() if stride == 1 else nn.AvgPool2d(kernel_size=stride, stride=stride)   # modification
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.pool(out)    # modification

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNetAP(ResNet):
    def __init__(self, *args, pool_size: int = 4, **kwargs):
        super().__init__(*args, **kwargs)
        self.avgpool = nn.Identity()    # modification
        self.fc = nn.Linear(self.fc.in_features * (pool_size * pool_size), self.fc.out_features)    # modification

    def _make_layer(self, block: type[BasicBlockAP | BottleneckAP], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride=1),  # modification
                nn.AvgPool2d(kernel_size=stride, stride=stride),    # modification
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)


def _resnet(
    block: type[BasicBlockAP | BottleneckAP],
    layers: list[int],
    **kwargs: Any
) -> ResNetAP:
    return ResNetAP(block, layers, **kwargs)


def resnet18(**kwargs: Any) -> ResNetAP:
    return _resnet(BasicBlockAP, [2, 2, 2, 2], **kwargs)



#!/usr/bin/env python3

import torch
import math

__all__ = ['tanh_func', 'atan_func',
           'repeat_to_batch', 'add_noise']


def tanh_func(x: torch.Tensor) -> torch.Tensor:
    r"""tanh object function.

    .. code-block:: python

        return x.tanh().add(1).mul(0.5)

    Args:
        x (torch.Tensor): The tensor ranging from
            :math:`[-\infty, +\infty]`.
    Returns:
        torch.Tensor: The tensor ranging in ``[0, 1]``
    """
    return x.tanh().add(1).mul(0.5)


def atan_func(x: torch.Tensor) -> torch.Tensor:
    r"""arctan object function.

    .. code-block:: python

        return x.atan().div(math.pi).add(0.5)

    Args:
        x (torch.Tensor): The tensor ranging from
            :math:`[-\infty], +\infty]`.
    Returns:
        torch.Tensor: The tensor ranging in ``[0, 1]``
    """
    return x.atan().div(math.pi).add(0.5)
# --------------------------------------------------------------------- #


def repeat_to_batch(x: torch.Tensor, batch_size: int = 1) -> torch.Tensor:
    r"""Repeat a single input tensor to a batch.

    Args:
        x (torch.Tensor): The single input tensor to process
            with shape ``(*)``.
        batch_size (int): Batch size. Defaults to ``1``.
    Returns:
        torch.Tensor:
            The batched input tensor with shape ``(batch_size, *)``
    """
    return x.expand([batch_size] + [-1] * x.dim())


def add_noise(x: torch.Tensor, noise: torch.Tensor = None,
              mean: float = 0.0, std: float = 1.0, universal: bool = False,
              clip_min: float | torch.Tensor = 0.0,
              clip_max: float | torch.Tensor = 1.0) -> torch.Tensor:
    r"""Add noise to a batched input tensor.

    Args:
        x (torch.Tensor): The input tensor to process
            with shape ``(N, *)``.
        noise (torch.Tensor | None): The pre-defined noise.
            If ``None``, generate Gaussian noise
            using :any:`torch.normal`.
            Defaults to be ``None``.
        mean (float): The mean of generated Gaussian noise.
            Defaults to ``0.0``.
        std (float): The std of generated Gaussian noise.
            Defaults to ``1.0``.
        universal (bool): Whether the noise is universal
            for all samples in the batch.
            Defaults to ``False``.
        clip_min (float | torch.Tensor):
            The min value of available input region.
            Defaults to ``0.0``.
        clip_max (float | torch.Tensor):
            The max value of available input region.
            Defaults to ``1.0``.

    Returns:
        torch.Tensor:
            The noisy batched input tensor
            with shape ``(N, *)`` (``(*)`` when ``universal=True``).
    """
    if noise is None:
        shape = x.shape
        if universal:
            shape = shape[1:]
        noise = torch.normal(mean=mean, std=std,
                             size=shape, device=x.device)
    batch_noise = noise
    if universal:
        batch_noise = repeat_to_batch(noise, x.shape[0])
    noisy_input: torch.Tensor = (
        x + batch_noise).clamp(clip_min, clip_max)
    return noisy_input



#!/usr/bin/env python3

from ...abstract import ModelInspection

import torch
import argparse


class NeuralCleanse(ModelInspection):
    r"""Neural Cleanse proposed by Bolun Wang and Ben Y. Zhao
    from University of Chicago in IEEE S&P 2019.

    It is a model inspection backdoor defense
    that inherits :class:`trojanvision.defenses.ModelInspection`.
    (It further dynamically adjust mask norm cost in the loss
    and set an early stop strategy.)

    For each class, Neural Cleanse tries to optimize a recovered trigger
    that any input with the trigger attached will be classified to that class.
    If there is an outlier among all potential triggers, it means the model is poisoned.

    See Also:
        * paper: `Neural Cleanse\: Identifying and Mitigating Backdoor Attacks in Neural Networks`_
        * code: https://github.com/bolunwang/backdoor

    Args:
        nc_cost_multiplier (float): Norm loss cost multiplier.
            Defaults to ``1.5``.
        nc_patience (float): Early stop nc_patience.
            Defaults to ``10.0``.
        nc_asr_threshold (float): ASR threshold in cost adjustment.
            Defaults to ``0.99``.
        nc_early_stop_threshold (float): Threshold in early stop check.
            Defaults to ``0.99``.

    Attributes:
        cost_multiplier_up (float): Value to multiply when increasing cost.
            It equals to ``nc_cost_multiplier``.
        cost_multiplier_down (float): Value to divide when decreasing cost.
            It's set as ``nc_cost_multiplier ** 1.5``.

    Attributes:
        init_cost (float): Initial cost of mask norm loss.
        cost (float): Current cost of mask norm loss.

    .. _Neural Cleanse\: Identifying and Mitigating Backdoor Attacks in Neural Networks:
        https://ieeexplore.ieee.org/document/8835365
    """
    name: str = 'neural_cleanse'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--nc_cost_multiplier', type=float,
                           help='norm loss cost multiplier '
                           '(default: 1.5)')
        group.add_argument('--nc_patience', type=float,
                           help='early stop nc_patience '
                           '(default: 10.0)')
        group.add_argument('--nc_asr_threshold', type=float,
                           help='asr threshold in cost adjustment '
                           '(default: 0.99)')
        group.add_argument('--nc_early_stop_threshold', type=float,
                           help='threshold in early stop check. '
                           '(default: 0.99)')
        return group

    def __init__(self, nc_cost_multiplier: float = 1.5, nc_patience: float = 10.0,
                 nc_asr_threshold: float = 0.99,
                 nc_early_stop_threshold: float = 0.99, **kwargs):
        super().__init__(**kwargs)
        self.init_cost = self.cost
        self.param_list['neural_cleanse'] = ['cost_multiplier_up', 'cost_multiplier_down',
                                             'nc_patience', 'nc_asr_threshold',
                                             'nc_early_stop_threshold']
        self.cost_multiplier_up = nc_cost_multiplier
        self.cost_multiplier_down = nc_cost_multiplier ** 1.5
        self.nc_asr_threshold = nc_asr_threshold
        self.nc_early_stop_threshold = nc_early_stop_threshold
        self.nc_patience = nc_patience
        self.early_stop_patience = self.nc_patience * 2

    def optimize_mark(self, *args, **kwargs) -> tuple[torch.Tensor, float]:
        # parameters to update cost
        self.cost_set_counter = 0
        self.cost_up_counter = 0
        self.cost_down_counter = 0
        self.cost_up_flag = False
        self.cost_down_flag = False

        # counter for early stop
        self.early_stop_counter = 0
        self.early_stop_norm_best = float('inf')
        return super().optimize_mark(*args, **kwargs)

    def check_early_stop(self, acc: float, norm: float, **kwargs) -> bool:
        # update cost
        if self.cost == 0 and acc >= self.nc_asr_threshold:
            self.cost_set_counter += 1
            if self.cost_set_counter >= self.nc_patience:
                self.cost = self.init_cost
                self.cost_up_counter = 0
                self.cost_down_counter = 0
                self.cost_up_flag = False
                self.cost_down_flag = False
                # print(f'initialize cost to {self.cost:.2f}%.2f')
        else:
            self.cost_set_counter = 0

        if acc >= self.nc_asr_threshold:
            self.cost_up_counter += 1
            self.cost_down_counter = 0
        else:
            self.cost_up_counter = 0
            self.cost_down_counter += 1

        if self.cost_up_counter >= self.nc_patience:
            self.cost_up_counter = 0
            # prints(f'up cost from {self.cost:.4f} to {self.cost * self.cost_multiplier_up:.4f}',
            #        indent=4)
            self.cost *= self.cost_multiplier_up
            self.cost_up_flag = True
        elif self.cost_down_counter >= self.nc_patience:
            self.cost_down_counter = 0
            # prints(f'down cost from {self.cost:.4f} to {self.cost / self.cost_multiplier_down:.4f}',
            #        indent=4)
            self.cost /= self.cost_multiplier_down
            self.cost_down_flag = True

        early_stop = False
        # check early stop
        if norm < float('inf'):
            if norm >= self.nc_early_stop_threshold * self.early_stop_norm_best:
                self.early_stop_counter += 1
            else:
                self.early_stop_counter = 0
        self.early_stop_norm_best = min(norm, self.early_stop_norm_best)

        if self.cost_down_flag and self.cost_up_flag and self.early_stop_counter >= self.early_stop_patience:
            early_stop = True

        return early_stop



#!/usr/bin/env python3

from trojanvision.models.imagemodel import ImageModel

from .alexnet import AlexNet
from .densenet import DenseNet
from .efficientnet import EfficientNet
from .mnasnet import MNASNet
from .mobilenet import MobileNet
from .resnet import ResNet
from .shufflenetv2 import ShuffleNetV2
from .vgg import VGG

__all__ = ['AlexNet', 'DenseNet', 'EfficientNet', 'MNASNet',
           'MobileNet', 'ResNet', 'ShuffleNetV2', 'VGG']

class_dict: dict[str, type[ImageModel]] = {
    'alexnet': AlexNet,
    'densenet': DenseNet,
    'efficientnet': EfficientNet,
    'mnasnet': MNASNet,
    'mobilenet': MobileNet,
    'resnet': ResNet,
    'resnext': ResNet,
    'shufflenetv2': ShuffleNetV2,
    'vgg': VGG,
}



#!/usr/bin/env python3

r"""It is equivalent to :ref:`trojanzoo.environ <trojanzoo.environ>`.

Note:
    The only difference is that it uses ``trojanvision.configs.config``
    as the default parameter passed to :func:`trojanzoo.environ.create()`.
"""

from trojanvision.configs import config
from trojanzoo.environ import add_argument, env  # noqa: F401
import trojanzoo.environ

from trojanvision.configs import Config


def create(config: Config = config, **kwargs) -> trojanzoo.environ.Env:
    return trojanzoo.environ.create(config=config, **kwargs)


create()
